@@ -79,7 +79,6 @@ def list_target_files(target_name, coredata, builddata):
     try:
         t = builddata.targets[target_name]
         sources = t.sources + t.extra_files
-        subdir = t.subdir
     except KeyError:
         print("Unknown target %s." % target_name)
         sys.exit(1)
@@ -87,30 +86,23 @@ def list_target_files(target_name, coredata, builddata):
     print(json.dumps(sources))
 
 def list_buildoptions(coredata, builddata):
-    buildtype= {'choices': ['plain', 'debug', 'debugoptimized', 'release'],
+    buildtype= {'choices': ['plain', 'debug', 'debugoptimized', 'release', 'minsize'],
                 'type' : 'combo',
-                'value' : coredata.buildtype,
+                'value' : coredata.get_builtin_option('buildtype'),
                 'description' : 'Build type',
                 'name' : 'type'}
-    strip = {'value' : coredata.strip,
+    strip = {'value' : coredata.get_builtin_option('strip'),
              'type' : 'boolean',
              'description' : 'Strip on install',
              'name' : 'strip'}
-    coverage = {'value': coredata.coverage,
-                'type' : 'boolean',
-                'description' : 'Enable coverage',
-                'name' : 'coverage'}
-    pch = {'value' : coredata.use_pch,
-           'type' : 'boolean',
-           'description' : 'Use precompiled headers',
-           'name' : 'pch'}
-    unity = {'value' : coredata.unity,
+    unity = {'value' : coredata.get_builtin_option('unity'),
              'type' : 'boolean',
              'description' : 'Unity build',
              'name' : 'unity'}
-    optlist = [buildtype, strip, coverage, pch, unity]
+    optlist = [buildtype, strip, unity]
     add_keys(optlist, coredata.user_options)
     add_keys(optlist, coredata.compiler_options)
+    add_keys(optlist, coredata.base_options)
     print(json.dumps(optlist))
 
 def add_keys(optlist, options):
@@ -121,14 +113,14 @@ def add_keys(optlist, options):
         optdict = {}
         optdict['name'] = key
         optdict['value'] = opt.value
-        if isinstance(opt, mesonlib.UserStringOption):
+        if isinstance(opt, coredata.UserStringOption):
             typestr = 'string'
-        elif isinstance(opt, mesonlib.UserBooleanOption):
+        elif isinstance(opt, coredata.UserBooleanOption):
             typestr = 'boolean'
-        elif isinstance(opt, mesonlib.UserComboOption):
+        elif isinstance(opt, coredata.UserComboOption):
             optdict['choices'] = opt.choices
             typestr = 'combo'
-        elif isinstance(opt, mesonlib.UserStringArrayOption):
+        elif isinstance(opt, coredata.UserStringArrayOption):
             typestr = 'stringarray'
         else:
             raise RuntimeError("Unknown option type")
