@@ -34,21 +34,22 @@ import concurrent.futures as conc
 from mesonbuild.coredata import backendlist
 
 class TestResult:
-    def __init__(self, msg, stdo, stde, conftime=0, buildtime=0, testtime=0):
+    def __init__(self, msg, stdo, stde, mlog, conftime=0, buildtime=0, testtime=0):
         self.msg = msg
         self.stdo = stdo
         self.stde = stde
+        self.mlog = mlog
         self.conftime = conftime
         self.buildtime = buildtime
         self.testtime = testtime
 
 class AutoDeletedDir():
-    def __init__(self, dir):
-        self.dir = dir
+    def __init__(self, d):
+        self.dir = d
     def __enter__(self):
         os.makedirs(self.dir, exist_ok=True)
         return self.dir
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, _type, value, traceback):
         # On Windows, shutil.rmtree fails sometimes, because 'the directory is not empty'.
         # Retrying fixes this.
         # That's why we don't use tempfile.TemporaryDirectory, but wrap the deletion in the AutoDeletedDir class.
@@ -65,6 +66,7 @@ class AutoDeletedDir():
 passing_tests = 0
 failing_tests = 0
 skipped_tests = 0
+failing_logs = []
 print_debug = 'MESON_PRINT_TEST_OUTPUT' in os.environ
 
 meson_command = os.path.join(os.getcwd(), 'meson')
@@ -114,52 +116,61 @@ def setup_commands(backend):
         backend_flags = []
         ninja_command = environment.detect_ninja()
         if ninja_command is None:
-            raise RuntimeError('Could not find Ninja executable.')
+            raise RuntimeError('Could not find Ninja v1.6 or newer')
         if print_debug:
             compile_commands = [ninja_command, '-v']
         else:
             compile_commands = [ninja_command]
+        compile_commands += ['-w', 'dupbuild=err']
         test_commands = [ninja_command, 'test', 'benchmark']
         install_commands = [ninja_command, 'install']
 
-def platform_fix_filename(fname):
-    if mesonlib.is_osx():
-        if fname.endswith('.so'):
-            return fname[:-2] + 'dylib'
-        return fname.replace('.so.', '.dylib.')
-    elif mesonlib.is_windows():
-        if fname.endswith('.so'):
-            (p, f) = os.path.split(fname)
-            f = f[3:-2] + 'dll'
-            return os.path.join(p, f)
-        if fname.endswith('.a'):
-            return fname[:-1] + 'lib'
+def get_relative_files_list_from_dir(fromdir):
+    paths = []
+    for (root, _, files) in os.walk(fromdir):
+        reldir = os.path.relpath(root, start=fromdir)
+        for f in files:
+            path = os.path.join(reldir, f).replace('\\', '/')
+            if path.startswith('./'):
+                path = path[2:]
+            paths.append(path)
+    return paths
+
+def platform_fix_exe_name(fname):
+    if not fname.endswith('?exe'):
+        return fname
+    fname = fname[:-4]
+    if mesonlib.is_windows():
+        return fname + '.exe'
     return fname
 
 def validate_install(srcdir, installdir):
-    if mesonlib.is_windows():
-        # Don't really know how Windows installs should work
-        # so skip.
-        return ''
+    # List of installed files
     info_file = os.path.join(srcdir, 'installed_files.txt')
+    # If this exists, the test does not install any other files
+    noinst_file = 'usr/no-installed-files'
     expected = {}
     found = {}
-    if os.path.exists(info_file):
+    ret_msg = ''
+    # Generate list of expected files
+    if os.path.exists(os.path.join(installdir, noinst_file)):
+        expected[noinst_file] = False
+    elif os.path.exists(info_file):
         for line in open(info_file):
-            expected[platform_fix_filename(line.strip())] = True
-    for root, _, files in os.walk(installdir):
-        for fname in files:
-            found_name = os.path.join(root, fname)[len(installdir)+1:]
-            found[found_name] = True
-    expected = set(expected)
-    found = set(found)
-    missing = expected - found
-    for fname in missing:
-        return 'Expected file %s missing.' % fname
-    extra = found - expected
-    for fname in extra:
-        return 'Found extra file %s.' % fname
-    return ''
+            expected[platform_fix_exe_name(line.strip())] = False
+    # Check if expected files were found
+    for fname in expected:
+        if os.path.exists(os.path.join(installdir, fname)):
+            expected[fname] = True
+    for (fname, found) in expected.items():
+        if not found:
+            ret_msg += 'Expected file {0} missing.\n'.format(fname)
+    # Check if there are any unexpected files
+    found = get_relative_files_list_from_dir(installdir)
+    for fname in found:
+        if fname not in expected:
+            ret_msg += 'Extra file {0} found.\n'.format(fname)
+    return ret_msg
 
 def log_text_file(logfile, testdir, stdo, stde):
     global stop, executor, futures
@@ -236,13 +247,18 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, flags, compile_c
     gen_command = [meson_command, '--prefix', '/usr', '--libdir', 'lib', testdir, test_build_dir]\
         + flags + test_args + extra_args
     (returncode, stdo, stde) = run_configure_inprocess(gen_command)
+    try:
+        logfile = os.path.join(test_build_dir, 'meson-logs/meson-log.txt')
+        mesonlog = open(logfile, errors='ignore').read()
+    except Exception:
+        mesonlog = 'No meson-log.txt found.'
     gen_time = time.time() - gen_start
     if not should_succeed:
         if returncode != 0:
-            return TestResult('', stdo, stde, gen_time)
-        return TestResult('Test that should have failed succeeded', stdo, stde, gen_time)
+            return TestResult('', stdo, stde, mesonlog, gen_time)
+        return TestResult('Test that should have failed succeeded', stdo, stde, mesonlog, gen_time)
     if returncode != 0:
-        return TestResult('Generating the build system failed.', stdo, stde, gen_time)
+        return TestResult('Generating the build system failed.', stdo, stde, mesonlog, gen_time)
     if 'msbuild' in compile_commands[0]:
         sln_name = glob(os.path.join(test_build_dir, '*.sln'))[0]
         comp = compile_commands + [os.path.split(sln_name)[-1]]
@@ -256,7 +272,7 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, flags, compile_c
     stdo += o.decode(sys.stdout.encoding)
     stde += e.decode(sys.stdout.encoding)
     if pc.returncode != 0:
-        return TestResult('Compiling source code failed.', stdo, stde, gen_time, build_time)
+        return TestResult('Compiling source code failed.', stdo, stde, mesonlog, gen_time, build_time)
     test_start = time.time()
     # Note that we don't test that running e.g. 'ninja test' actually
     # works. One hopes that this is a common enough happening that
@@ -266,7 +282,7 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, flags, compile_c
     stdo += tstdo
     stde += tstde
     if returncode != 0:
-        return TestResult('Running unit tests failed.', stdo, stde, gen_time, build_time, test_time)
+        return TestResult('Running unit tests failed.', stdo, stde, mesonlog, gen_time, build_time, test_time)
     if len(install_commands) == 0:
         return TestResult('', '', '', gen_time, build_time, test_time)
     else:
@@ -278,8 +294,8 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, flags, compile_c
         stdo += o.decode(sys.stdout.encoding)
         stde += e.decode(sys.stdout.encoding)
         if pi.returncode != 0:
-            return TestResult('Running install failed.', stdo, stde, gen_time, build_time, test_time)
-        return TestResult(validate_install(testdir, install_dir), stdo, stde, gen_time, build_time, test_time)
+            return TestResult('Running install failed.', stdo, stde, mesonlog, gen_time, build_time, test_time)
+        return TestResult(validate_install(testdir, install_dir), stdo, stde, mesonlog, gen_time, build_time, test_time)
 
 def gather_tests(testdir):
     tests = [t.replace('\\', '/').split('/', 2)[2] for t in glob(os.path.join(testdir, '*'))]
@@ -292,7 +308,7 @@ def detect_tests_to_run():
     all_tests = []
     all_tests.append(('common', gather_tests('test cases/common'), False))
     all_tests.append(('failing', gather_tests('test cases/failing'), False))
-    all_tests.append(('prebuilt object', gather_tests('test cases/prebuilt object'), False))
+    all_tests.append(('prebuilt', gather_tests('test cases/prebuilt'), False))
 
     all_tests.append(('platform-osx', gather_tests('test cases/osx'), False if mesonlib.is_osx() else True))
     all_tests.append(('platform-windows', gather_tests('test cases/windows'), False if mesonlib.is_windows() else True))
@@ -348,6 +364,7 @@ def run_tests(extra_args):
                     print('Failed test%s: %s' % (without_install, t))
                     print('Reason:', result.msg)
                     failing_tests += 1
+                    failing_logs.append(result.mlog)
                 else:
                     print('Succeeded test%s: %s' % (without_install, t))
                     passing_tests += 1
@@ -388,27 +405,51 @@ def check_format():
                 fullname = os.path.join(root, file)
                 check_file(fullname)
 
-def generate_prebuilt_object():
-    source = 'test cases/prebuilt object/1 basic/source.c'
-    objectbase = 'test cases/prebuilt object/1 basic/prebuilt.'
-    if shutil.which('cl'):
-        objectfile = objectbase + 'obj'
-        cmd = ['cl', '/nologo', '/Fo'+objectfile, '/c', source]
+def pbcompile(compiler, source, objectfile):
+    if compiler == 'cl':
+        cmd = [compiler, '/nologo', '/Fo'+objectfile, '/c', source]
     else:
-        if mesonlib.is_windows():
-            objectfile = objectbase + 'obj'
-        else:
-            objectfile = objectbase + 'o'
-        if shutil.which('cc'):
-            cmd = 'cc'
-        elif shutil.which('gcc'):
-            cmd = 'gcc'
-        else:
-            raise RuntimeError("Could not find C compiler.")
-        cmd = [cmd, '-c', source, '-o', objectfile]
-    subprocess.check_call(cmd)
+        cmd = [compiler, '-c', source, '-o', objectfile]
+    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
+
+def generate_pb_object(compiler, object_suffix):
+    source = 'test cases/prebuilt/1 object/source.c'
+    objectfile = 'test cases/prebuilt/1 object/prebuilt.' + object_suffix
+    pbcompile(compiler, source, objectfile)
     return objectfile
 
+def generate_pb_static(compiler, object_suffix, static_suffix):
+    source = 'test cases/prebuilt/2 static/libdir/best.c'
+    objectfile = 'test cases/prebuilt/2 static/libdir/best.' + object_suffix
+    stlibfile = 'test cases/prebuilt/2 static/libdir/libbest.' + static_suffix
+    pbcompile(compiler, source, objectfile)
+    if compiler == 'cl':
+        linker = ['lib', '/NOLOGO', '/OUT:' + stlibfile, objectfile]
+    else:
+        linker = ['ar', 'csr', stlibfile, objectfile]
+    subprocess.check_call(linker)
+    os.unlink(objectfile)
+    return stlibfile
+
+def generate_prebuilt():
+    static_suffix = 'a'
+    if shutil.which('cl'):
+        compiler = 'cl'
+        static_suffix = 'lib'
+    elif shutil.which('cc'):
+        compiler = 'cc'
+    elif shutil.which('gcc'):
+        compiler = 'gcc'
+    else:
+        raise RuntimeError("Could not find C compiler.")
+    if mesonlib.is_windows():
+        object_suffix = 'obj'
+    else:
+        object_suffix = 'o'
+    objectfile = generate_pb_object(compiler, object_suffix)
+    stlibfile = generate_pb_static(compiler, object_suffix, static_suffix)
+    return (objectfile, stlibfile)
+
 if __name__ == '__main__':
     parser = argparse.ArgumentParser(description="Run the test suite of Meson.")
     parser.add_argument('extra_args', nargs='*',
@@ -422,17 +463,19 @@ if __name__ == '__main__':
     if script_dir != '':
         os.chdir(script_dir)
     check_format()
-    pbfile = generate_prebuilt_object()
+    pbfiles = generate_prebuilt()
     try:
         run_tests(options.extra_args)
     except StopException:
         pass
-    os.unlink(pbfile)
+    for f in pbfiles:
+        os.unlink(f)
     print('\nTotal passed tests:', passing_tests)
     print('Total failed tests:', failing_tests)
     print('Total skipped tests:', skipped_tests)
-    if failing_tests > 0 and 'TRAVIS' in os.environ:
-        # Cat because it can have stuff of unknown encodings mixed.
-        subprocess.call(['cat', 'meson-test-run.txt'])
+    if failing_tests > 0 and ('TRAVIS' in os.environ or 'APPVEYOR' in os.environ):
+        print('\nMesonlogs of failing tests\n')
+        for l in failing_logs:
+            print(l, '\n')
     sys.exit(failing_tests)
 
